/*
** $Id: biniolib.c,v 0.1 13.07.2008 $
** Library for reading and writing binary files
** See Copyright Notice in lua.h
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>     /* for ftruncate and close */
#include <fcntl.h>      /* for open */
#include <sys/stat.h>   /* S_I* constants */
#include "config.h"   /* for Endianness */

#define biniolib_c
#define LUA_LIB

#include "agena.h"
#include "aauxlib.h"
#include "agenalib.h"

#include "chelpers.h"

#include "agncmpt.h"


static int binio_open (lua_State *L) {
  int hnd;
  const char *file = luaL_checkstring(L, 1);
  /* no option given: open in readwrite mode, option given: open in readonly mode */
  hnd = (lua_gettop(L) == 1) ? my_open(file) : my_roopen(file);
  if (hnd == -1) {
    if (agn_isverbose(L)) fprintf(stderr, "Error, file not found.\n");
    lua_pushfail(L);
    return 1;
  }
  /* enter new open file to global binio.openfiles table */
  lua_getglobal(L, "binio");
  lua_getfield(L, -1, "openfiles");
  lua_pushinteger(L, hnd);
  lua_pushstring(L, file);
  lua_rawset(L, -3);
  lua_pop(L, 2);  /* delete "binio" and "openfiles" */
  lua_pushinteger(L, hnd);
  return 1;
}


static int binio_close (lua_State *L) {
  int nargs, hnd, i, result;
  result = 1;
  nargs = lua_gettop(L);
  luaL_checkstack(L, nargs, "too many arguments");
  for (i=1; i <= nargs; i++) {
    hnd = luaL_checkinteger(L, i);
    if (hnd < 3) {  /* avoid standard streams to be closed */
      result = 0;
      continue;
    }
    if (!my_close(hnd) == 0)
      result = 0;
    else {
      /* delete file from global binio.openfiles table */
      lua_getglobal(L, "binio");
      lua_getfield(L, -1, "openfiles");
      lua_pushinteger(L, hnd);
      lua_pushnil(L);
      lua_rawset(L, -3);
      lua_pop(L, 2);  /* delete "binio" and "openfiles" */
    }
  }
  if (result)
    lua_pushtrue(L);
  else
    lua_pushfail(L);
  return 1;
}


static int binio_readchar (lua_State *L) {
  long pos;
  int hnd, res;
  unsigned char data;
  hnd = luaL_checkinteger(L, 1);
  pos = luaL_optinteger(L, 2, 0);
  if (pos != 0 && lseek(hnd, pos, SEEK_CUR) == -1) {
    lua_pushfail(L);
    return 1;
  }
  res = read(hnd, &data, sizeof(char));
  if (res)
    lua_pushnumber(L, data);
  else if (res == 0)
    lua_pushnil(L);
  else
    lua_pushfail(L);
  return 1;
}


/* base_sync: flushes all unwritten content to the database file. */

static int binio_sync (lua_State *L) {
  int hnd, try, res;
  hnd = luaL_checkinteger(L, 1);
  /* file is not open ? -> leave this here since there will be lock errors later with closed files*/
  try = my_seek(hnd, 0L);
  if (try == -1) {
    lua_pushfail(L);
    return 1;
  }
  /* lock the file, seeker is at position 0 already */
  my_lock(hnd, 4L);
  #ifdef _WIN32
  res = _commit(hnd);
  #else
  res = fsync(hnd);
  #endif
  /* unlock file */
  my_seek(hnd, 0L);
  my_unlock(hnd, 4L);
  if (!res)
    lua_pushboolean(L, 1);
  else
    lua_pushfail(L);
  return 1;
}


static int binio_filepos (lua_State *L) {
  int hnd;
  long fpos;
  hnd = luaL_checkinteger(L, 1);
  fpos = my_fpos(hnd);
  if (fpos == -1)
    lua_pushfail(L);
  else
    lua_pushnumber(L, fpos);
  return 1;
}


static int binio_seek (lua_State *L) {
  int hnd;
  long pos;
  hnd = luaL_checkinteger(L, 1);
  pos = luaL_checkinteger(L, 2);
  if (lseek(hnd, pos, SEEK_CUR) == -1)
    lua_pushfail(L);
  else
    lua_pushtrue(L);
  return 1;
}


static int binio_length (lua_State *L) {
  int hnd;
  long size, oldpos;
  hnd = luaL_checkinteger(L, 1);
  oldpos = lseek(hnd, 0L, SEEK_CUR);
  size = lseek(hnd, 0L, SEEK_END);
  lseek(hnd, oldpos, SEEK_SET);  /* reset cursor to original file position */
  if (size == -1)
    lua_pushfail(L);
  else
    lua_pushnumber(L, size);
  return 1;
}


static int binio_make (lua_State *L) {
  int hnd;
  const char *file = luaL_checkstring(L, 1);
  /* my_create in chelpers.c does not work for it unexplainably creates a read-only file
     whereas in the ads package it does not */
  hnd = open(file, O_BINARY|O_RDWR|O_CREAT|O_TRUNC, S_IREAD|S_IWRITE);
  if (hnd < 0)
    lua_pushfail(L);
  else
    lua_pushinteger(L, hnd);
  return 1;  /* the file remains open */
}


static int binio_writechar (lua_State *L) {
  int hnd;
  size_t size;
  unsigned char data;
  hnd = luaL_checkinteger(L, 1);
  data = agn_checknumber(L, 2);
  size = sizeof(data);
  if (write(hnd, &data, size) != size)
    lua_pushfail(L);
  else
    lua_pushtrue(L);
  return 1;
}


static int binio_writenumber (lua_State *L) {
  int hnd;
  size_t size;
  lua_Number data;
  hnd = luaL_checkinteger(L, 1);
  data = agn_checknumber(L, 2);
  #if BYTE_ORDER != BIG_ENDIAN
  tools_swapdouble(&data);
  #endif
  size = sizeof(data);
  if (write(hnd, &data, size) != size)
    lua_pushfail(L);
  else
    lua_pushtrue(L);
  return 1;
}


static int binio_writelong (lua_State *L) {
  int hnd;
  size_t size;
  long data;
  hnd = luaL_checkinteger(L, 1);
  data = agn_checknumber(L, 2);
  #if BYTE_ORDER != BIG_ENDIAN
  //data = tools_swaplong(data);
  tools_swaplong(&data);
  #endif
  size = sizeof(data);
  if (write(hnd, &data, size) != size)
    lua_pushfail(L);
  else
    lua_pushtrue(L);
  return 1;
}


static int binio_writestring (lua_State *L) {
  int hnd;
  size_t l;
  const char *value;
  hnd = luaL_checkinteger(L, 1);
  value = agn_checklstring(L, 2, &l);
  char data[l];
  strcpy(data, value);
  my_writel(hnd, l);
  if (write(hnd, &data, l) != l)
    lua_pushfail(L);
  else
    lua_pushtrue(L);
  return 1;
}


static int binio_readstring (lua_State *L) {
  int hnd, success;
  size_t size;
  hnd = luaL_checkinteger(L, 1);
  size = sec_readl(hnd, &success);  /* first read length of the string */
  if (success) {
    char data[size+1];
    if (read(hnd, &data, size) == size) {  /* now read string itself */
      data[size] = '\0';
      lua_pushlstring(L, data, size);
      return 1;
    }
  }
  lua_pushfail(L);
  return 1;
}


static int binio_readnumber (lua_State *L) {
  int hnd;
  size_t success;
  lua_Number data;
  hnd = luaL_checkinteger(L, 1);
  success = sec_read(hnd, &data, sizeof(lua_Number));
  if (success) {
    #if BYTE_ORDER != BIG_ENDIAN
    tools_swapdouble(&data);
    #endif
    lua_pushnumber(L, data);
    return 1;
  }
  lua_pushfail(L);
  return 1;
}


static int binio_readlong (lua_State *L) {
  int hnd;
  size_t success;
  long data;
  hnd = luaL_checkinteger(L, 1);
  success = sec_read(hnd, &data, sizeof(long));
  if (success) {
    #if BYTE_ORDER != BIG_ENDIAN
    tools_swaplong(&data);
    lua_pushnumber(L, data);
    #else
    lua_pushnumber(L, data);
    #endif
    return 1;
  }
  lua_pushfail(L);
  return 1;
}


static int binio_toend (lua_State *L) {
  int hnd;
  long size;
  hnd = luaL_checkinteger(L, 1);
  size = lseek(hnd, 0L, SEEK_END);
  if (size == -1)
    lua_pushfail(L);
  else
    lua_pushnumber(L, size);
  return 1;
}


static int binio_rewind (lua_State *L) {
  int hnd;
  long size;
  hnd = luaL_checkinteger(L, 1);
  size = lseek(hnd, 0L, SEEK_SET);
  if (size == -1)
    lua_pushfail(L);
  else
    lua_pushnumber(L, size);
  return 1;
}


static const luaL_Reg binio[] = {
  {"open", binio_open},                    /* July 13, 2008 */
  {"close", binio_close},                  /* July 13, 2008 */
  {"readchar", binio_readchar},            /* July 13, 2008 */
  {"sync", binio_sync},                    /* July 13, 2008 */
  {"filepos", binio_filepos},              /* July 13, 2008 */
  {"seek", binio_seek},                    /* July 13, 2008 */
  {"length", binio_length},                /* July 13, 2008 */
  {"make", binio_make},                    /* July 19, 2008 */
  {"writechar", binio_writechar},          /* July 19, 2008 */
  {"writestring", binio_writestring},      /* July 20, 2008 */
  {"writenumber", binio_writenumber},      /* October 25, 2008 */
  {"writelong", binio_writelong},          /* October 26, 2008 */
  {"readstring", binio_readstring},        /* July 20, 2008 */
  {"readnumber", binio_readnumber},        /* October 25, 2008 */
  {"readlong", binio_readlong},            /* October 26, 2008 */
  {"rewind", binio_rewind},                /* July 20, 2008 */
  {"toend", binio_toend},                  /* July 20, 2008 */
  {NULL, NULL}
};


static void createmetatable (lua_State *L) {
  lua_createtable(L, 0, 1);  /* create metatable for strings */
  lua_pushliteral(L, "");  /* dummy string */
  lua_pushvalue(L, -2);
  lua_setmetatable(L, -2);  /* set string metatable */
  agn_poptop(L);  /* pop dummy string */
  lua_pushvalue(L, -2);  /* string library... */
  lua_setfield(L, -2, "__index");  /* ...is the __index metamethod */
  agn_poptop(L);  /* pop metatable */
}


/*
** Open binio library
*/
LUALIB_API int luaopen_binio (lua_State *L) {
  luaL_register(L, AGENA_BINIOLIBNAME, binio);
  lua_newtable(L);
  lua_setfield(L, -2, "openfiles");  /* table for information on all open files */
  createmetatable(L);
  return 1;
}


