/*
** $Id: lmathlib.c,v 1.67 2005/08/26 17:36:32 roberto Exp $
** Standard mathematical library
** See Copyright Notice in lua.h
*/


#include <stdlib.h>
#include <math.h>

#define lmathlib_c
#define LUA_LIB

#include "agena.h"

#include "aauxlib.h"
#include "agenalib.h"

#include "agncmpt.h"  /* for trunc and isfinite */


#undef PHI
#define PHI      ((1.0+sqrt(5.0))/2.0)


/* much faster than an Agena implementation using (exp(x)-exp(-x))/2 */

static int math_sinh (lua_State *L) {
  switch (lua_type(L, 1)) {
    case LUA_TNUMBER: lua_pushnumber(L, sinh(agn_tonumber(L, 1))); break;
#ifndef LUA_ANSI
    case LUA_TCOMPLEX: agn_createcomplex(L, csinh(agn_tocomplex(L, 1))); break;
#else
    case LUA_TCOMPLEX: {
      lua_Number a = agn_complexreal(L, 1), b = agn_compleximag(L, 1);
      agn_createcomplex(L, sinh(a)*cos(b), cosh(a)*sin(b)); break;
    }
#endif
    default: luaL_error(L, "in " LUA_QS ", number or complex value expected.", "math.sinh");
  }
  return 1;
}


static int math_cosh (lua_State *L) {
  switch (lua_type(L, 1)) {
    case LUA_TNUMBER: lua_pushnumber(L, cosh(agn_tonumber(L, 1))); break;
#ifndef LUA_ANSI
    case LUA_TCOMPLEX: agn_createcomplex(L, ccosh(agn_tocomplex(L, 1))); break;
#else
    case LUA_TCOMPLEX: {
      lua_Number a = agn_complexreal(L, 1), b = agn_compleximag(L, 1);
      agn_createcomplex(L, cosh(a)*cos(b), sinh(a)*sin(b)); break;
    }
#endif
    default: luaL_error(L, "in " LUA_QS ", number or complex value expected.", "math.cosh");
  }
  return 1;
}


static int math_tanh (lua_State *L) {
  switch (lua_type(L, 1)) {
    case LUA_TNUMBER: lua_pushnumber(L, tanh(agn_tonumber(L, 1))); break;
#ifndef LUA_ANSI
    case LUA_TCOMPLEX: agn_createcomplex(L, ctanh(agn_tocomplex(L, 1))); break;
#else
    case LUA_TCOMPLEX: {
      lua_Number t1, t4, t5, t6, t8;
      lua_Number a = agn_complexreal(L, 1), b = agn_compleximag(L, 1);
      t1 = sinh(a);
      t4 = t1*t1;
      t5 = cos(b);
      t6 = t5*t5;
      t8 = 1/(t4+t6);
      agn_createcomplex(L, t1*cosh(a)*t8, sin(b)*t5*t8);
      break;
    }
#endif
    default: luaL_error(L, "in " LUA_QS ", number or complex value expected.", "math.tanh");
  }
  return 1;
}


static int math_arcsin (lua_State *L) {
  switch (lua_type(L, 1)) {
    case LUA_TNUMBER: {
      lua_Number x = agn_tonumber(L, 1);
      if (x < -1 || x > 1)
        lua_pushundefined(L);
      else
        lua_pushnumber(L, asin(x));
      break;
    }
#ifndef LUA_ANSI
    case LUA_TCOMPLEX: agn_createcomplex(L, casin(agn_tocomplex(L, 1))); break;
#else
    case LUA_TCOMPLEX: {
      lua_Number t1, t2, t4, t6, t12, t14;
      lua_Number a = agn_complexreal(L, 1), b = agn_compleximag(L, 1);
      t1 = a*a;
      t2 = b*b;
      t4 = sqrt(t1+2.0*a+1.0+t2);
      t6 = sqrt(t1-2.0*a+1.0+t2);
      t12 = pow(t4/2+t6/2,2.0);
      t14 = sqrt(t12-1.0);
      agn_createcomplex(L, asin(t4/2-t6/2), csgn(b, -a)*log(t4/2+t6/2+t14));
      break;
    }
#endif
    default: luaL_error(L, "in " LUA_QS ", number or complex value expected.", "math.arcsin");
  }
  return 1;
}


static int math_arccos (lua_State *L) {
  switch (lua_type(L, 1)) {
    case LUA_TNUMBER: {
      lua_Number x = agn_tonumber(L, 1);
      if (x < -1 || x > 1)
        lua_pushundefined(L);
      else
        lua_pushnumber(L, acos(x));
      break;
    }
#ifndef LUA_ANSI
    case LUA_TCOMPLEX: agn_createcomplex(L, cacos(agn_tocomplex(L, 1))); break;
#else
    case LUA_TCOMPLEX: {
      lua_Number t1, t2, t4, t6, t12, t14;
      lua_Number a = agn_complexreal(L, 1), b = agn_compleximag(L, 1);
      t1 = a*a;
      t2 = b*b;
      t4 = sqrt(t1+2.0*a+1.0+t2);
      t6 = sqrt(t1-2.0*a+1.0+t2);
      t12 = pow(t4/2+t6/2,2.0);
      t14 = sqrt(t12-1.0);
      agn_createcomplex(L, acos(t4/2-t6/2), csgn(-b, a)*log(t4/2+t6/2+t14));
      break;
    }
#endif
    default: luaL_error(L, "in " LUA_QS ", number or complex value expected.", "math.arccos");
  }
  return 1;
}


static int math_arctan2 (lua_State *L) {
  lua_pushnumber(L, atan2(agn_checknumber(L, 1), agn_checknumber(L, 2)));
  return 1;
}


static int math_arccoth (lua_State *L) {
  switch (lua_type(L, 1)) {
    case LUA_TNUMBER: {
      lua_Number x, arg;
      x = agn_tonumber(L, 1);
      arg = (x+1.0)/(x-1.0);
      if (arg <= 0 || isnan(arg) || !isfinite(arg))
        lua_pushundefined(L);
      else
        lua_pushnumber(L, 0.5*log(arg));
      break;
    }
    case LUA_TCOMPLEX: {
      lua_Number re, im, t1, t2, t3, t6;
#ifndef LUA_ANSI
      agn_Complex z = agn_tocomplex(L, 1);
      re = creal(z); im = cimag(z);
#else
      re = agn_complexreal(L, 1); im = agn_compleximag(L, 1);
#endif
      t1 = re+1;
      t2 = t1*t1;
      t3 = im*im;
      t6 = pow(re-1, 2);
#ifndef LUA_ANSI
      agn_createcomplex(L, log((t2+t3)/(t6+t3))/4+I*(PI/2+atan2(im, t1)/2-atan2(-im, 1-re)/2));
#else
      agn_createcomplex(L, log((t2+t3)/(t6+t3))/4, PI/2+atan2(im, t1)/2-atan2(-im, 1-re)/2);
#endif
      break;
    }
    default: luaL_error(L, "in " LUA_QS ", number or complex value expected.", "math.arccoth");
  }
  return 1;
}


static int math_argument (lua_State *L) {
  switch (lua_type(L, 1)) {
    case LUA_TCOMPLEX:
#ifndef LUA_ANSI
      lua_pushnumber(L, carg(agn_tocomplex(L, 1))); break;
#else
      { lua_Number a = agn_complexreal(L, 1), b = agn_compleximag(L, 1);
      lua_pushnumber(L, atan2(b, a)); break; }
#endif
    case LUA_TNUMBER:
      lua_pushnumber(L, 0); break;
    default: luaL_error(L, "in " LUA_QS ", number or complex value expected.", "math.argument");
  }
  return 1;
}


/* 9 % faster than the Agena implementation */

static int math_conjugate (lua_State *L) {
  switch (lua_type(L, 1)) {
    case LUA_TCOMPLEX:
#ifndef LUA_ANSI
      agn_createcomplex(L, conj(agn_tocomplex(L, 1))); break;
#else
      { lua_Number a = agn_complexreal(L, 1), b = agn_compleximag(L, 1);
      agn_createcomplex(L, a, -b); break; }
#endif
    case LUA_TNUMBER:
      lua_pushvalue(L, 1);
      break;
    default: luaL_error(L, "in " LUA_QS ", number or complex value expected.", "math.conjugate");
  }
  return 1;
}


static int math_root (lua_State *L) {
  lua_Number n;
  luaL_checkany(L, 2);
  n = agn_checknumber(L, 2);
  if (trunc(n) != n)
    luaL_error(L, "second argument must be an integer, got " LUA_NUMBER_FMT, n);
  switch (lua_type(L, 1)) {
    case LUA_TNUMBER: {
      lua_Number x = agn_tonumber(L, 1);
      if (x >= 0)
        lua_pushnumber(L, pow(x, 1/n));
      else if (fmod(n, 2) == 0)  /* x is even ? */
        lua_pushundefined(L);
      else
        lua_pushnumber(L, -pow(-x, 1/n));
      break;
    }
#ifndef LUA_ANSI
    case LUA_TCOMPLEX: {
      agn_Complex x = agn_tocomplex(L, 1);
      lua_Number k = trunc(carg(x)*(n-1)/PI/2);
      if (trunc(k) == k)
        agn_createcomplex(L, cpow(x, 1/n)*exp(2*I*k*PI/n));
      else
        lua_pushfail(L);
      break;
    }
#else
    case LUA_TCOMPLEX: {
      lua_Number t1, t2, t3, t7, t9;
      lua_Number a = agn_complexreal(L, 1), b = agn_compleximag(L, 1);
      t1 = 1/n;
      t2 = a*a;
      t3 = b*b;
      t7 = exp(t1*log(t2+t3)/2);
      t9 = t1*atan2(b,a);
      agn_createcomplex(L, t7*cos(t9), t7*sin(t9));
      break;
    }
#endif
    default: luaL_error(L, "in " LUA_QS ", number or complex value expected.", "math.root");
  }
  return 1;
}


static int math_fmod (lua_State *L) {
  lua_pushnumber(L, fmod(agn_checknumber(L, 1), agn_checknumber(L, 2)));
  return 1;
}


static int math_modf (lua_State *L) {
  lua_Number ip;
  lua_Number fp = modf(agn_checknumber(L, 1), &ip);
  lua_pushnumber(L, ip);
  lua_pushnumber(L, fp);
  return 2;
}


static int math_frexp (lua_State *L) {
  int e;
  lua_pushnumber(L, frexp(agn_checknumber(L, 1), &e));
  lua_pushinteger(L, e);
  return 2;
}


static int math_ldexp (lua_State *L) {
  lua_pushnumber(L, ldexp(agn_checknumber(L, 1), luaL_checkint(L, 2)));
  return 1;
}


static int math_min (lua_State *L) {
  int n = lua_gettop(L);  /* number of arguments */
  lua_Number dmin = agn_checknumber(L, 1);
  int i;
  for (i=2; i<=n; i++) {
    lua_Number d = agn_checknumber(L, i);
    if (d < dmin)
      dmin = d;
  }
  lua_pushnumber(L, dmin);
  return 1;
}


static int math_max (lua_State *L) {
  int n = lua_gettop(L);  /* number of arguments */
  lua_Number dmax = agn_checknumber(L, 1);
  int i;
  for (i=2; i<=n; i++) {
    lua_Number d = agn_checknumber(L, i);
    if (d > dmax)
      dmax = d;
  }
  lua_pushnumber(L, dmax);
  return 1;
}


static int math_random (lua_State *L) {
  /* the `%' avoids the (rare) case of r==1, and is needed also because on
     some systems (SunOS!) `rand()' may return a value larger than RAND_MAX */
  lua_Number r = (lua_Number)(rand()%RAND_MAX) / (lua_Number)RAND_MAX;
  switch (lua_gettop(L)) {  /* check number of arguments */
    case 0: {  /* no arguments */
      lua_pushnumber(L, r);  /* Number between 0 and 1 */
      break;
    }
    case 1: {  /* only upper limit */
      int u = luaL_checkint(L, 1);
      luaL_argcheck(L, 1<=u, 1, "interval is empty");
      lua_pushnumber(L, floor(r*u)+1);  /* int between 1 and `u' */
      break;
    }
    case 2: {  /* lower and upper limits */
      int l = luaL_checkint(L, 1);
      int u = luaL_checkint(L, 2);
      luaL_argcheck(L, l<=u, 2, "interval is empty");
      lua_pushnumber(L, floor(r*(u-l+1))+l);  /* int between `l' and `u' */
      break;
    }
    default: return luaL_error(L, "in " LUA_QS ", wrong number of arguments.", "math.random");
  }
  return 1;
}


static int math_randomseed (lua_State *L) {
  srand(luaL_checkint(L, 1));
  return 0;
}


/* math_roundf: rounds real value x to the d-th digit; patched Dec 22, 2007 for negative values */

static int math_roundf (lua_State *L) {
  lua_Number x = agn_checknumber(L, 1);
  lua_Number d = luaL_optinteger(L, 2, 0);
  if (x < 0)
    lua_pushnumber(L, ceil(pow(10, d)*x-0.5) * pow(10, (-d)));
  else
    lua_pushnumber(L, floor(pow(10, d)*x+0.5) * pow(10, (-d)));
  return 1;
}


/* isfloat: check whether a Lua number is a float; April 15, 2007
   ! The function is 9 % faster than an Agena implementation */

static int math_isfloat (lua_State *L) {
  lua_Number x = agn_checknumber(L, 1);
  lua_pushboolean(L, (trunc(x) != x));
  return 1;
}


/* isinteger: check whether a Lua number is an integer; June 29, 2008
   ! The function is 25 % faster than an Agena implementation */

static int math_isinteger (lua_State *L) {
  lua_Number x = agn_checknumber(L, 1);
  lua_pushboolean(L, (trunc(x) == x));  /* do not try (x == (long int)x), since this fails */
  return 1;
}


/* remainder of an integer division; June 19, 2007 */

static int math_irem (lua_State *L) {
  div_t result;
  result = div(agn_checknumber(L, 1), agn_checknumber(L, 2));
  lua_pushnumber(L, result.rem);
  return 1;
}


/* hypotenuse; June 19, 2007 */

static int math_hypot (lua_State *L) {
  lua_pushnumber(L, hypot(agn_checknumber(L, 1), agn_checknumber(L, 2)));
  return 1;
}


static int math_isprime (lua_State *L) {
  lua_pushboolean(L, tools_isPrime(luaL_checkint(L, 1)));
  return 1;
}


/* Added December 22, 2007; 0.9.0; much faster than an Agena implementation.
   Function value table added June 30, 2008 */

/* first 101 factorials created with Maple V Release 4
  for i from 0 to 100 by 3 do
     printf(`  %0.16e, %0.16e, %0.16e,\n`, factorial(i), factorial(i+1), factorial(i+2))
  od:
*/
static double factorials[] = {
  1.0000000000000000e+00, 1.0000000000000000e+00, 2.0000000000000000e+00,
  6.0000000000000000e+00, 2.4000000000000000e+01, 1.2000000000000000e+02,
  7.2000000000000000e+02, 5.0400000000000000e+03, 4.0320000000000000e+04,
  3.6288000000000000e+05, 3.6288000000000000e+06, 3.9916800000000000e+07,
  4.7900160000000000e+08, 6.2270208000000000e+09, 8.7178291200000000e+10,
  1.3076743680000000e+12, 2.0922789888000000e+13, 3.5568742809600000e+14,
  6.4023737057280000e+15, 1.2164510040883200e+17, 2.4329020081766400e+18,
  5.1090942171709440e+19, 1.1240007277776076e+21, 2.5852016738884976e+22,
  6.2044840173323943e+23, 1.5511210043330985e+25, 4.0329146112660563e+26,
  1.0888869450418352e+28, 3.0488834461171386e+29, 8.8417619937397019e+30,
  2.6525285981219105e+32, 8.2228386541779228e+33, 2.6313083693369353e+35,
  8.6833176188118864e+36, 2.9523279903960414e+38, 1.0333147966386144e+40,
  3.7199332678990121e+41, 1.3763753091226345e+43, 5.2302261746660111e+44,
  2.0397882081197443e+46, 8.1591528324789773e+47, 3.3452526613163807e+49,
  1.4050061177528798e+51, 6.0415263063373835e+52, 2.6582715747884487e+54,
  1.1962222086548019e+56, 5.5026221598120889e+57, 2.5862324151116818e+59,
  1.2413915592536072e+61, 6.0828186403426756e+62, 3.0414093201713378e+64,
  1.5511187532873822e+66, 8.0658175170943878e+67, 4.2748832840600255e+69,
  2.3084369733924138e+71, 1.2696403353658275e+73, 7.1099858780486345e+74,
  4.0526919504877216e+76, 2.3505613312828785e+78, 1.3868311854568983e+80,
  8.3209871127413901e+81, 5.0758021387722479e+83, 3.1469973260387937e+85,
  1.9826083154044400e+87, 1.2688693218588416e+89, 8.2476505920824706e+90,
  5.4434493907744306e+92, 3.6471110918188685e+94, 2.4800355424368305e+96,
  1.7112245242814131e+98, 1.1978571669969891e+100, 8.5047858856786231e+101,
  6.1234458376886086e+103, 4.4701154615126843e+105, 3.3078854415193864e+107,
  2.4809140811395398e+109, 1.8854947016660502e+111, 1.4518309202828586e+113,
  1.1324281178206297e+115, 8.9461821307829752e+116, 7.1569457046263802e+118,
  5.7971260207473679e+120, 4.7536433370128417e+122, 3.9455239697206586e+124,
  3.3142401345653532e+126, 2.8171041143805502e+128, 2.4227095383672732e+130,
  2.1077572983795277e+132, 1.8548264225739843e+134, 1.6507955160908461e+136,
  1.4857159644817614e+138, 1.3520015276784029e+140, 1.2438414054641307e+142,
  1.1567725070816415e+144, 1.0873661566567430e+146, 1.0329978488239059e+148,
  9.9167793487094968e+149, 9.6192759682482119e+151, 9.4268904488832477e+153,
  9.3326215443944152e+155, 9.3326215443944152e+157, 9.4259477598383594e+159
};


static int math_fact (lua_State *L) {
  lua_Number x = agn_checknumber(L, 1);
  if (trunc(x) == x && x > -1)  /* is x an integer and nonnegative ? */
    if (x < 101)
      lua_pushnumber(L, factorials[(int)x]);
    else
      lua_pushnumber(L, exp(lgamma(x+1)));
  else
    lua_pushundefined(L);
  return 1;
}


/* Added December 26, 2007; 0.9.0; changed April 20, 2008 - 0.10.0 */

/* computes the binomial coefficient */

static int math_binomial (lua_State *L) {
  lua_Number n = agn_checknumber(L, 1);
  lua_Number k = agn_checknumber(L, 2);
  if (n < 0 || k < 0) /* || trunc(n) != n || trunc(k) != k || n < k) */
    lua_pushundefined(L);
  else
    lua_pushnumber(L, exp(lgamma(n+1) - lgamma(k+1) - lgamma(n-k+1)));
  return 1;
}


/* constant found in Cephes Math Library Release 2.8, June, 2000 by Stephen L. Moshier
   file sin.c (and fixed it) */
#define P64800   (4.8481368110953599358991410e-6)  /* (Pi/180)/3600 */

/* converts degrees to radians */
static int math_toRadians (lua_State *L) {
  lua_Number m, s;
  m = luaL_optnumber(L, 2, 0);
  s = luaL_optnumber(L, 3, 0);
  lua_pushnumber(L,
    ((agn_checknumber(L, 1)*60.0 + m)*60.0 + s)*P64800
  );
  return 1;
}


static const luaL_Reg mathlib[] = {
  {"arccos",  math_arccos},
  {"arcsin",  math_arcsin},
  {"arctan2", math_arctan2},
  {"binomial", math_binomial},            /* added on December 26, 2007 */
  {"cosh", math_cosh},
  {"fact", math_fact},                    /* added on December 22, 2007 */
  {"fmod", math_fmod},
  {"frexp", math_frexp},
  {"hypot", math_hypot},                  /* added on June 19, 2007 */
  {"irem", math_irem},                    /* added on June 19, 2007 */
  {"isfloat", math_isfloat},              /* added on April 15, 2007 */
  {"isinteger", math_isinteger},          /* added on June 29, 2008 */
  {"isprime", math_isprime},              /* added on August 19, 2007 */
  {"ldexp", math_ldexp},
  {"max", math_max},
  {"min", math_min},
  {"modf", math_modf},
  {"random", math_random},
  {"randomseed", math_randomseed},
  {"roundf", math_roundf},                /* added on December 03, 2006 */
  {"sinh", math_sinh},
  {"tanh", math_tanh},
  {"toRadians", math_toRadians},          /* added on January 21, 2008 */
  {"argument", math_argument},            /* added on June 29, 2008 */
  {"conjugate", math_conjugate},          /* added on June 29, 2008 */
  {"root", math_root},                    /* added on June 29, 2008 */
  {"arccoth", math_arccoth},              /* added on June 29, 2008 */
  {NULL, NULL}
};


/*
** Open math library
*/
LUALIB_API int luaopen_math (lua_State *L) {
  luaL_register(L, LUA_MATHLIBNAME, mathlib);
  lua_pushnumber(L, PHI);      /* added 0.5.0 */
  lua_setfield(L, -2, "Phi");  /* added 0.5.0 */
/* #if defined(LUA_COMPAT_MOD) */
  lua_getfield(L, -1, "fmod");
  lua_setfield(L, -2, "mod");
/* #endif */
  return 1;
}


